/*global window document $ localStorage alert*/

// UI.js (c) 2010-2013 Loren West and other contributors
// May be freely distributed under the MIT license.
// For further details and documentation:
// http://github.com/lorenwest/monitor-dashboard
(function(root){

  // Module loading
  var Monitor = root.Monitor || require('monitor'),
      OS = Monitor.commonJS ? require('os') : null,
      Backbone = Monitor.Backbone,
      _ = Monitor._,
      templates = {};  // Cache key=app-name, value=Template object

  // Constants
  var KEY_ESC = 27;

  /**
  * Node Monitor Namespace
  *
  * This class is here to provide a global namespace to attach all UI classes
  * and UI utility functions.
  *
  * @static
  * @class UI
  * @constructor
  */
  var UI = Monitor.UI = function(){};
  var Mustache = UI.Mustache = Monitor.commonJS ? require('../ext/mustache-0.7.0-dev.js') : root.Mustache;

  /**
  * Load the specified template
  *
  * This loads the template from cache, from the DOM (if present), or from the
  * server if it's not in DOM.
  *
  * It can be run on the client side only
  *
  * @static
  * @method loadTemplate
  * @param app {String} Application name
  * @param name {String} Template name
  * @param callback {Function(error, template)} Called when the template is loaded or on error.
  *     @param callback.error {Object} Object containing the error or null if no error
  *     @param callback.template {Template} The Template object
  */
  UI.loadTemplate = function(app, name, callback) {

    // See if the template is in cache
    var cacheKey = app ? app + '-' + name : name;
    if (templates[cacheKey]) {
      return callback(null, templates[cacheKey]);
    }

    // See if the template is in DOM
    var domId = '#nm-template-' + cacheKey;
    if ($(domId).length) {
      return callback(null, templates[cacheKey] = new UI.Template({
        text: $(domId).html()
      }));
    }

    // Get the template from the server
    var url = (app ? '/app/' + app : '/static') + '/template/' + name + '.html';
    $.get(url, function(data){
      // If we have a <doctype>, then we've gotten a 404
      if (data.indexOf('<!DOCTYPE html>') === 0) {
        return callback({code:'ENOENT', msg:'No template found at ' + url});
      }
      templates[cacheKey] = new UI.Template({text:data});
      return callback(null, templates[cacheKey]);
    }, 'html').error(callback);
  };

  /**
  * Static Configurations
  *
  * These can be set onto the UI class after it's loaded, or via the
  * node.js config package if running server-side.
  *
  * @static
  * @property Config
  * @type &lt;Object&gt;</code>
  * <ul>
  * </ul>
  */
  var defaultConfig = {
    port: 4200,
    allowExternalConnections: false,
    siteDbPath: './site_db'
  };
  if (Monitor.commonJS) {
    Monitor.Config.setModuleDefaults('Dashboard', defaultConfig);
  } else {
    _.extend(Monitor.Config, {Monitor: defaultConfig});
  }
  UI.Config = Monitor.Config.Monitor;

  // Create the app namespace
  var app = UI.app = {};

  // Expose jquery if on the client.  No longer needed on the server.
  var $ = UI.$ = root.$ || {fn:{}};

  /* jQuery no longer needed on the server
  // Set up server-side jQuery
  $.support.cors = true;
  if (Monitor.commonJS) {
    var XHR = require('xmlhttprequest').XMLHttpRequest;
    $.ajaxSetup({xhr: function(){return new XHR();}});
  }
  */

  // A generic mask to use for modals.  Can be stacked for modal within modal
  // (example: dropdown within a dialog).
  UI.ModalMask = Backbone.View.extend({
    initialize: function(options){
      var t = this;

      // Render on new
      t.$el.addClass('nm-mask').appendTo('body');

      // Attaching keydown to the document vs. the view
      t.keydown = t.onKeyDown.bind(t);
      $(document).on('keydown', t.keydown);

    },
    events: {
      'click'   : 'close'
    },
    onKeyDown: function(e) {
      var t = this;
      if (e.keyCode === KEY_ESC) {
        return t.close(e);
      }
    },
    close: function(e) {
      var t = this;
      t.trigger('close');
      t.$el.remove();
      $(document).off('keydown', t.keydown);
      e.stopPropagation();
      e.preventDefault();
      return false;
    }
  });

  // Common dialog box move logic
  UI.modalDragStart = function(e) {
    var dialog = $(e.currentTarget).parent(),
        zoomRatio = window.innerWidth / window.outerWidth,
        newPosition = {
          top: parseInt(dialog.css('top'), 10) * zoomRatio,
          left: parseInt(dialog.css('left'), 10) * zoomRatio
        },
        offset = {
          top: e.pageY - newPosition.top,
          left: e.pageX - newPosition.left
        };

    function drag(e) {
      newPosition.top = Math.max(e.pageY - offset.top, 0);
      newPosition.left = Math.max(e.pageX - offset.left, 0);
      dialog.css(newPosition);
    }
    function drop(e) {
      dialog.addClass('fade');
      $(document).unbind("mousemove", drag).unbind("mouseup", drop);
    }
    $(document).bind("mousemove", drag).bind("mouseup", drop);
    dialog.removeClass('fade');
    drag(e);
    e.preventDefault();
  };

  /**
  * Merge changes from another array or collection
  *
  * This adds a ```set()``` method to Backbone.Collection, which merges the
  * specified array or collection into the existing collection.
  *
  * It is useful for streaming updates into existing collections without
  * disrupting any event listeners attached to existing models.
  *
  * If an item already exists, it merges changes into it.  If an item
  * doesn't exist it adds it, and it removes any item in the collection that
  * isn't in the array or collection passed in to the method.
  *
  * Example:
  *
  *     var stream = JSON.stringify(myCollection);
  *     var myArray = JSON.parse(stream);
  *     myArray[0].lastUpdated = new Date();
  *     myCollection.set(myArray);
  *
  * @method Backbone.Collection.set
  * @param items {Array or Backbone.Collection} The items to set into the collection
  * @param options {Object} Set options to pass to individual model create/set
  */
  Backbone.Collection.prototype.set = function(items, options) {

    var t = this,
        prevModels = t.models,
        newIds = items.pluck ? items.pluck('id') : _.pluck(items, 'id'),
        i, len, id, cid, prevModel, newItem, removeIds = [], newIdHash = {};

    // Merge or add items to this collection
    for (i = 0, len = newIds.length; i < len; i++) {
      id = newIds[i];
      newItem = items[i];
      newIdHash[id] = true;
      prevModel = t.get(id);
      if (prevModel) {
        prevModel.set(newItem instanceof Backbone.Model ? newItem.attributes : newItem, options);
      } else {
        t.add(newItem, options);
      }
    }

    // Remove any item not in new items
    t.each(function(model) {
      id = model.get('id');
      if (!newIdHash[id]) {
        removeIds.push(id);
      }
    });
    t.remove(removeIds);
  };

  /**
  * Set up a contained model or contained collection relationship
  *
  * This is called during initialize of a parent backbone model that contains
  * another backbone model or collection.  It:
  *
  *     * Creates the correct model if setting from a non-model object
  *     * Merges changes on set vs. overwriting the existing contained object
  *     * Merges collection items vs. replacing (if it's a collection)
  *     * Alters toJSON to return the contained object toJSON vs. the object reference
  *
  * Example:
  *
  *     UI.containedModel(this, 'author', Author);
  *
  * @method containedModel
  * @param model {Backbone.Model} The parent model object
  * @param attrName {String} Name of the attribute that contains the model or collection
  * @param Ctor {Function} The constructor function of the contained model or collection
  */
  UI.containedModel = function(model, attrName, Ctor) {

    // Build the function for setting model data
    function setAttr(setModel, newValue, setOptions) {
      var oldValue = model._containedModels[attrName];

      // Pass through if removing
      if (newValue === undefined || newValue === null) {
        model._containedModels[attrName] = newValue;
        return;
      }

      // Is the new value the correct type?
      if (newValue instanceof Ctor) {

        // Directly set if no old value
        if (!oldValue instanceof Ctor) {
          model._containedModels[attrName] = newValue;
          return;
        }

        // They're both models.  Disregard if they're the same.
        var oldJSON = oldValue.toJSON({deep:true});
        var newJSON = newValue.toJSON({deep:true});
        if (_.isEqual(oldJSON, newJSON)) {
          return;
        }

        // Merge the raw JSON if they're both models
        newValue = newJSON;
      }

      // Keep the previous model and merge new data into it
      // For collections this relies on the Collection.set() method
      if (oldValue instanceof Ctor) {
        model.attributes[attrName] = oldValue;
        model._currentAttributes[attrName] = oldValue;
        model.changed[attrName] = oldValue;
        oldValue.set(newValue, setOptions);
        return;
      }

      // Create a new model or collection, passing the value
      newValue =
        model._containedModels[attrName] =
        model.attributes[attrName] =
        model._currentAttributes[attrName] =
        new Ctor(newValue, setOptions);

      // Watch for changes to the underlying model or collection
      // (collection add/remove), forwarding changes to this model
      newValue.on('change add remove reset', UI.onSubModelChange, {
        parent:model,
        attrName: attrName
      });
    }

    // Remember the contained models before they're overridden by set
    if (!model._containedModels) {
      model._containedModels = {};
    }

    // Watch for changes to the underlying model or collection
    // (collection add/remove), forwarding changes to this model
    var subModel = model.get(attrName);
    if (subModel instanceof Backbone.Model || subModel instanceof Backbone.Collection) {
      subModel.on('change add remove reset', UI.onSubModelChange, {
        parent:model,
        attrName: attrName
      });
    }

    // Set the initial model data
    setAttr(model, model.get(attrName));

    // Watch for subsequent changes
    model.on('change:' + attrName, setAttr, model);
  };

  /**
  * Change listener attached to contained models - for bubbling change
  * events to the containing model.
  *
  * @method onSubModelChange
  * @param model {Model} Backbone model changed
  * @param [collection] {Backbone collection} if a collection add/remove
  * @param options {Object} Event options
  * @param this {Object} 'this' is a custom object containing:
  *     @param this.parent {Backbone.Model} Parent container for this object
  *     @param this.attrName {String} Parents attribute name for this model
  */
  UI.onSubModelChange = function(model) {
    var t = this,
        parent = t.parent,
        attrName = t.attrName,
        // Options is arg[2] for add/remove and arg[1] for change
        chgOptions = arguments[2] || arguments[1] || {};

    // If the parent isn't currently changing, the source of the change
    // came from the contained model.  Trigger changes to the parent model.
    if (!chgOptions.silent && !parent._changing) {

      // Use setTimeout to bundle up all changes into a single change event
      // for each attribute and the parent class.
      parent._pendingChanges = parent._pendingChanges || {};
      parent._pendingChanges[attrName] = true;
      if (!parent._subModelTimer) {

        // Create one timer for all changes to the parent
        parent._subModelTimer = setTimeout(function(){
          var changes = {};
          for (var chgAttrName in parent._pendingChanges) {

            // Get the changed attribute for the set
            changes[chgAttrName] = parent.attributes[chgAttrName];

            // Make sure a change is detected
            parent.attributes[chgAttrName] = '';
            parent._currentAttributes[chgAttrName] = '';
            parent._previousAttributes[chgAttrName] = '';
          }

          // Set all sub-model changes into the parent
          parent.set(changes);

          // Remove our internal members
          delete parent._pendingChanges;
          delete parent._subModelTimer;
        },0);
      }
    } else {

      // Inform the parent that this attribute has pending changes.
      // This is necessary for the parent to fire it's own change event.
      parent._pending[attrName] = true;
    }
  };

  /**
  * Additional options for the Backbone.js toJSON method
  *
  * If the options hash contains {deep:true}, then call the toJSON() method of
  * any contained objects that have a toJSON method.
  *
  * If the options hash contains {trim:true}, then only output values that
  * differ from the default value.  This produces a less chatty JSON string,
  * useful for for streaming models across the wire.
  *
  * @method toJSON
  * @return {Object} A raw JS copy of the data model
  */
  Backbone.Model.prototype.toJSON = function(options) {
    var t = this,
        result = {},
        defaults = _.isFunction(t.defaults) ? t.defaults() : t.defaults,
        attrs = t.attributes,
        attr, value;

    options = options || {};

    for (attr in attrs) {
      value = attrs[attr];
      if (options.deep && value && _.isFunction(value.toJSON)) {
        value = value.toJSON(options);
      }
      if (options.trim && defaults) {
        if (!_.isEqual(value, defaults[attr])) {
          result[attr] = value;
        }
      } else {
        result[attr] = value;
      }
    }
    return result;
  };

  /**
  * Attach a common tooltip to the specified selector
  *
  * This attaches a bootstrap.js tooltip onto the elements specified with
  * the jQuery selector.  It adds classes to the tooltip elements, allowing them
  * to be individually styled.
  *
  * @method tooltip
  * @param selector {jQuery Selector} Selector of items to add tooltips onto
  * @param options {Object} Tool tip options
  */
  UI.tooltip = function(selector, options) {
    selector.addClass('nm-pv-tt').tooltip(_.extend({
      placement: 'left',
      template: '<div class="tooltip nm-pv-tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>',
      delay: {show: 500, hide:10}
    }, options));
  };

  /**
  * Hide all tooltips and cancel awaiting timers
  *
  * @method hideToolTips
  */
  UI.hideToolTips = function() {
    $('.nm-pv-tt').each(function() {
      var $el = $(this),
          tooltip = $el.data('tooltip');
      if (tooltip) {
        clearTimeout($el.data('tooltip').timeout);
      }
      $el.tooltip('hide');
    });
  };

  /**
  * Pause the tour if running
  */
  UI.pauseTour = function() {
    if (UI.pageView.tourView) {
      UI.pageView.tourView.pause();
    }
  };

  /**
  * jQuery add-on for inserting CSS into the DOM
  *
  * This method can accept a CSS style sheet in a string, an array of strings,
  * or an object with keys representing selectors and values representing styles.
  *
  * String example:
  *
  *     $.styleSheet("body: {background:white;}");
  *
  * Array example:
  *
  *     $.styleSheet([
  *       "body: {background:white;}",
  *       "#canvas: background: #e0e0e0; max-width: 400px;"
  *     ]);
  *
  * Object example:
  *
  *     $.styleSheet({
  *       "body"    : "background: white;",
  *       "#canvas" : "background: #e0e0e0; max-width: 400px;"
  *     });
  *
  * Deep object example:
  *
  *     $.styleSheet({
  *       "body": {
  *         "background": "white",
  *       },
  *       "#canvas":
  *         "background": "#e0e0e0",
  *         "max-width": "400px";"
  *     });
  *
  * @static
  * method $.styleSheet
  * @param styles {String, Array, or Object} Styles to apply to the page
  * @param id {String} ID of the style element.  This places an ID attribute onto the style tag, and replaces any prior style element with the specified ID.
  */
  $.styleSheet = function(styles, id) {

    // Build the style string based on the input type
    var styleStr = '', idAttr = '';
    if (typeof styles === 'string') {
      styleStr = styles;
    }
    else if (Array.isArray(styles)) {
      styleStr = styles.join('\n');
    }
    else if (typeof styles === 'object') {
      for (var selector in styles) {
        var value = styles[selector], valueStr = '';
        if (typeof value === 'string') {
          valueStr = value;
        }
        else if (typeof value === 'object') {
          for (var cssAttr in value) {
            valueStr += cssAttr + ':' + value[cssAttr] + "; ";
          }
        }
        styleStr += selector + "{" + valueStr + "} ";
      }
    }

    // Remove any prior style with the specified ID
    if (id) {
      $('#' + id).remove();
      idAttr = 'id="' + id + '" ';
    }

    // Insert into DOM
    if (styleStr) {
      $('<style ' + idAttr + 'type="text/css">' + styleStr + '</style>').appendTo('head');
    }
  };

  /**
  * jQuery utility for centering a div on the screen
  *
  * @static
  * @method $.centerBox
  */
  $.fn.centerBox = function () {
    var t = this;
    t.css("top", Math.max(0, (($(window).height() - this.outerHeight()) / 2) + $(window).scrollTop()) + "px");
    t.css("left", Math.max(0, (($(window).width() - this.outerWidth()) / 2) + $(window).scrollLeft()) + "px");
    return t;
  };

  /**
  * jQuery utility for converting a style string into an object
  *
  * This converts a string like this:
  *
  *     var style = "top:15px; left:200px; background-color:#202020";
  *
  * Into an object like this:
  *
  *     {
  *       "top": "15px",
  *       "left": "200px",
  *       "background-color": "#202020"
  *     }
  *
  * @static
  * method $.parseStyleString
  * @param styleString {String} CSS Style string
  * @return {Object} The parsed styles, as an object
  */
  $.parseStyleString = function(styleString) {
    var parsed = {}, parts = styleString.split(';');
    for (var i = 0, l = parts.length; i < l; i++) {
      var nameVal = parts[i].split(':');
      if (nameVal.length === 2) {
        parsed[nameVal[0].trim()] = nameVal[1].trim();
      }
    }
    return parsed;
  };

  /**
  * jQuery utility for converting a CSS object into a style string
  *
  * This converts an object like this:
  *
  *     var css = {
  *       "top": "15px",
  *       "left": "200px",
  *       "background-color": "#202020"
  *     }
  *
  * Into a string like this:
  *
  *     "top:15px; left:200px; background-color:#202020;"
  *
  * @static
  * method $.makeStyleString
  * @param css {Object} The CSS object
  * @return {String} The style string
  */
  $.makeStyleString = function(css) {
    var styles = '', separator = '';
    for (var elemName in css) {
      styles += separator + elemName + ":" + css[elemName] + ';';
      separator = ' ';
    }
    return styles;
  };

  /**
  * jQuery utility for making a string Title Case
  *
  * @static
  * method $.titleCase
  * @param titleString {String} A title string
  * @param preserveCase {String} Don't lowercase the string first
  * @return {String} The titleString Title Cased
  */
  $.titleCase = function(titleString, preserveCase) {
    var newTitle = preserveCase ? titleString : titleString.toLowerCase();
    newTitle = newTitle.split(' ');
    for (var i = 0, l = newTitle.length; i < l; i++) {
      newTitle[i] = newTitle[i].substr(0,1).toUpperCase() + newTitle[i].substr(1);
    }
    return newTitle.join(' ');
  };

  // Add Function.bind to browsers that don't support it yet (mobile safari & ie).
  // Thanks to https://developer.mozilla.org/en/JavaScript/Reference/Global_Objects/Function/bind
  if (!Function.prototype.bind) {
    Function.prototype.bind = function (oThis) {
      if (typeof this !== "function") {
        // closest thing possible to the ECMAScript 5 internal IsCallable function
        throw new TypeError("Function.prototype.bind - what is trying to be bound is not callable");
      }
      var aArgs = Array.prototype.slice.call(arguments, 1),
        fToBind = this,
        FNOP = function () {},
        fBound = function () {
          return fToBind.apply(this instanceof FNOP ? this : oThis, aArgs.concat(Array.prototype.slice.call(arguments)));
        };
      FNOP.prototype = this.prototype;
      fBound.prototype = new FNOP();
      return fBound;
    };
  }

}(this));
